from enum import Enum
from random import Random
from typing import Callable

from sampo.generator import SimpleSynthetic
from sampo.generator.types import SyntheticGraphType
from sampo.scheduler.utils.block_graph import BlockGraph
from sampo.scheduler.utils.obstruction import Obstruction


class SyntheticBlockGraphType(Enum):
    Sequential = 0,
    Parallel = 1,
    Random = 2,
    Queues = 3


def generate_blocks(graph_type: SyntheticBlockGraphType, n_blocks: int, type_prop: list[int],
                    count_supplier: Callable[[int], tuple[int, int]],
                    edge_prob: float, rand: Random | None = Random(),
                    obstruction_getter: Callable[[int], Obstruction | None] = lambda _: None) -> BlockGraph:
    """
    Generate synthetic block graph according to given parameters.

    :param graph_type: type of BlockGraph
    :param n_blocks: the count of blocks
    :param type_prop: proportions of the `WorkGraph` types: General, Parallel, Sequential
    :param count_supplier: function that computes the borders of block size from it's index
    :param edge_prob: edge existence probability
    :param rand: a random reference
    :param obstruction_getter:
    :return: generated block graph
    """
    ss = SimpleSynthetic(rand)

    modes = rand.sample(list(SyntheticGraphType), counts=[p * n_blocks for p in type_prop], k=n_blocks)
    nodes = [ss.work_graph(mode, *count_supplier(i)) for i, mode in enumerate(modes)]
    bg = BlockGraph(nodes, obstruction_getter)

    match graph_type:
        case SyntheticBlockGraphType.Sequential:
            for idx, start in enumerate(bg.nodes[:-2]):
                bg.add_edge(start, bg.nodes[idx + 1])
        case SyntheticBlockGraphType.Parallel:
            for idx, start in enumerate(bg.nodes[:-2]):
                bg.add_edge(start, bg.nodes[idx + 1])
        case SyntheticBlockGraphType.Random:
            rev_edge_prob = int(1 / edge_prob)
            for idx, start in enumerate(bg.nodes):
                for end in bg.nodes[idx:]:
                    if start == end or rand.randint(0, rev_edge_prob) != 0:
                        continue
                    bg.add_edge(start, end)

    return bg


from copy import deepcopy
from operator import attrgetter
from random import Random
from typing import Callable

from sampo.generator import SimpleSynthetic
from sampo.generator.types import SyntheticGraphType
from sampo.scheduler.utils.obstruction import Obstruction
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.utilities.collections import build_index


class BlockNode:
    """
    `BlockNode` represents the node of `BlockGraph` and contains corresponding
    `WorkGraph` and related dependencies between blocks in `BlockGraph`.
    """
    def __init__(self, wg: WorkGraph, obstruction: Obstruction | None = None):
        self.wg = wg
        self.obstruction = obstruction
        self.blocks_from: list[BlockNode] = []
        self.blocks_to: list[BlockNode] = []

    @property
    def id(self):
        return self.wg.start.id


class BlockGraph:
    """
    Represents the block graph, where blocks are instances of `WorkGraph` and
    edges are simple *FS* dependencies.
    """
    def __init__(self, nodes: list[WorkGraph], obstruction_getter: Callable[[int], Obstruction | None] = lambda _: None):
        self.nodes = [BlockNode(node, obstruction_getter(i)) for i, node in enumerate(nodes)]
        self.node_dict = build_index(self.nodes, attrgetter('id'))

    def __getitem__(self, item) -> BlockNode:
        return self.node_dict[item]

    def to_work_graph(self) -> WorkGraph:
        """
        Creates `WorkGraph` that are equal to this `BlockGraph`.
        """
        copied_nodes = deepcopy(self.nodes)
        global_start: GraphNode = [node for node in copied_nodes if len(node.blocks_from) == 0][0].wg.start
        global_end:   GraphNode = [node for node in copied_nodes if len(node.blocks_to) == 0][0].wg.finish

        for end in copied_nodes:
            end.wg.start.add_parents([start.wg.finish for start in end.blocks_from])

        return WorkGraph(global_start, global_end)

    @staticmethod
    def add_edge(start: BlockNode, end: BlockNode):
        start.blocks_to.append(end)
        end.blocks_from.append(start)



from dataclasses import dataclass
from typing import Dict

from sampo.scheduler.base import Scheduler
from sampo.scheduler.timeline.base import Timeline
from sampo.scheduler.utils.block_graph import BlockGraph
from sampo.scheduler.utils.obstruction import Obstruction
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph
from sampo.schemas.schedule import Schedule
from sampo.schemas.time import Time


class Agent:

    def __init__(self, name: str, scheduler: Scheduler, contractors: list[Contractor]):
        self.name = name
        self._timeline = None
        self._scheduler = scheduler
        self._contractors = contractors

    def offer(self, wg: WorkGraph) -> tuple[Time, Time, Schedule, Timeline]:
        """
        Computes the offer from agent to manager. Handles all works from given wg.

        :param wg: the given block of tasks
        :return: offered start time, end time, resulting schedule and timeline before offering

        To apply returned offer, use `Agent#confirm`.
        """
        schedule, start_time, timeline = \
            self._scheduler.schedule_with_cache(wg, self._contractors, timeline=self._timeline)
        return start_time, start_time + schedule.execution_time, schedule, timeline

    def confirm(self, timeline: Timeline):
        """
        Applies the given offer.

        :param timeline: timeline returned from corresponding `Agent#offer`
        """
        self._timeline = timeline

    def __str__(self):
        return f'Agent(name={self.name}, scheduler={self._scheduler.scheduler_type.name})'

    @property
    def contractors(self):
        return self._contractors


@dataclass
class ScheduledBlock:
    """
    An object represents scheduled graph block(group of works).

    Contains all data used in scheduling, the agent and resulting information.
    """
    wg: WorkGraph
    schedule: Schedule
    agent: Agent
    start_time: Time
    end_time: Time

    @property
    def id(self):
        return self.wg.start.id

    @property
    def duration(self):
        return self.end_time - self.start_time

    def __str__(self):
        return f'ScheduledBlock(start_time={self.start_time}, end_time={self.end_time}, agent={self.agent})'

    def __repr__(self):
        return str(self)


class Manager:
    def __init__(self, agents: list[Agent]):
        if len(agents) == 0:
            raise Exception("Manager can't work with empty list of agents")
        self._agents = agents

    # TODO Upgrade to supply the best parallelism
    def manage_blocks(self, bg: BlockGraph, log: bool = False) -> Dict[str, ScheduledBlock]:
        """
        Runs multi-agent system based on auction on given BlockGraph.
        
        :param bg: 
        :param log:
        :return: an index of resulting `ScheduledBlock`s built by ids of corresponding `WorkGraph`s
        """
        id2sblock = {}
        for i, block in enumerate(bg.nodes):
            if log:
                print('--------------------------------')
                print(f'Running auction on block {i}')
            agent_start_time, agent_end_time, agent_schedule, agent \
                = self.run_auction_with_obstructions(block.wg, block.obstruction)
            if log:
                print(f'Won agent {agent} with start time {agent_start_time} and end time {agent_end_time}, '
                      f'adding to scheduling history')
            max_parent_time = max((id2sblock[parent.id].end_time for parent in block.blocks_from), default=Time(0))
            start_time = max(max_parent_time, agent_start_time)
            delta = start_time - agent_start_time
            sblock = ScheduledBlock(wg=block.wg, agent=agent, schedule=agent_schedule,
                                    start_time=agent_start_time + delta,
                                    end_time=agent_end_time + delta)
            id2sblock[sblock.id] = sblock

        return id2sblock

    def run_auction_with_obstructions(self, wg: WorkGraph, obstruction: Obstruction | None):
        if obstruction:
            obstruction.generate(wg)
        return self.run_auction(wg)

    def run_auction(self, wg: WorkGraph) -> (Time, Time, Schedule, Agent):
        """
        Runs the auction on the given `WorkGraph`.

        :param wg: target `WorkGraph`
        :return: best start time, end time and the agent that is able to support this working time
        """
        best_start_time = 0
        best_end_time = Time.inf()
        best_schedule = None
        best_timeline = None
        best_agent = None

        offers = [(agent, agent.offer(wg)) for agent in self._agents]

        for offered_agent, (offered_start_time, offered_end_time, offered_schedule, offered_timeline) in offers:
            if offered_end_time - offered_start_time < best_end_time - best_start_time:
                best_start_time = offered_start_time
                best_end_time = offered_end_time
                best_schedule = offered_schedule
                best_timeline = offered_timeline
                best_agent = offered_agent
        best_agent.confirm(best_timeline)

        return best_start_time, best_end_time, best_schedule, best_agent
